@@ -131,6 +131,7 @@ __all__ = [
     "ExtendedUnbinnedNLL",
     "Template",
     "LeastSquares",
+    "NormalConstraint",
 ]
 
 T = TypeVar("T", float, NDArray)
@@ -1413,7 +1414,7 @@ class BinnedCost(MaskedCostWithPulls):
             cx = np.arange(len(n)).astype(float)
         else:
             xe = self.xe
-            cx = 0.5 * (xe[1:] + xe[:-1])
+            cx = 0.5 * (xe[1:] + xe[:-1])  # type: ignore
         plt.errorbar(cx, n, ne, fmt="ok")
         plt.stairs(mu, xe, fill=True, color="C0")
 
@@ -1874,7 +1875,7 @@ class Template(BinnedCost):
             cx = np.arange(len(n)).astype(float)
         else:
             xe = self.xe
-            cx = 0.5 * (xe[1:] + xe[:-1])
+            cx = 0.5 * (xe[1:] + xe[:-1])  # type: ignore
 
         plt.errorbar(cx, n, ne, fmt="ok")
 
@@ -2305,17 +2306,21 @@ class LeastSquares(MaskedCostWithPulls):
 
         x, y, ye = self._masked.T
         plt.errorbar(x, y, ye, fmt="ok")
+
+        xmin = np.min(x)
+        xmax = np.max(x)
         if isinstance(model_points, Iterable):
             xm = np.array(model_points)
             ym = self.model(xm, *args)
         elif model_points > 0:
+            # beware, x may not be sorted
             if _detect_log_spacing(x):
-                xm = np.geomspace(x[0], x[-1], model_points)
+                xm = np.geomspace(xmin, xmax, model_points)
             else:
-                xm = np.linspace(x[0], x[-1], model_points)
+                xm = np.linspace(xmin, xmax, model_points)
             ym = self.model(xm, *args)
         else:
-            xm, ym = _smart_sampling(lambda x: self.model(x, *args), x[0], x[-1])
+            xm, ym = _smart_sampling(lambda x: self.model(x, *args), xmin, xmax)
         plt.plot(xm, ym)
         return (x, y, ye), (xm, ym)
 
